//pl0中间代码生成函数的验证
// 要点：
//     1 pl0程序分成编译器和解释器两个部分，本验证文件只考虑编译器部分，解释器的验证暂不考虑。
//     2 因此，本文件只有1个函数：函数gen把三个参数x、y、z组装成一条目标指令并存放于code数组中，增加CX的值。
// 关键字：变量不变式、宏定义
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
// pl/0 compiler with code generation
#include <stdbool.h>
//#include <stdio.h>

#define norw		11		// no. of reserved words
#define txmax		100		// length of identifier table
#define nmax		9		// max. no. of digits in numbers
#define al		10		// length of identifiers
#define amax		2047		// maximum of constant value
#define levmax		3		// maximum depth of block nesting
#define cxmax		2000		// size of code array
#define dxmax		500		// maximum of variable address
#define errmax		300		// maximum of errors

#define nul		0x1		// 1
#define ident		0x2		// 2
#define number		0x4		// 4
#define plus		0x8		// 8
#define minus		0x10		// 16
#define times		0x20		// 32
#define slash		0x40		// 64
#define oddsym		0x80		// 128
#define eql		0x100		// 256
#define neq		0x200		// 512
#define lss		0x400		// 1024
#define leq		0x800		// 2048
#define gtr		0x1000		// 4096
#define geq		0x2000		// 8192
#define lparen		0x4000		// 16384
#define rparen		0x8000		// 32768
#define comma		0x10000		// 65536
#define semicolon	0x20000		// 131072
#define period		0x40000		// 262144
#define becomes		0x80000		// 524288
#define beginsym	0x100000	// 1048576
#define endsym		0x200000	// 2097152
#define ifsym		0x400000	// 4194304
#define thensym		0x800000	// 8388608
#define whilesym	0x1000000	// 16777216
#define dosym		0x2000000	// 33554432
#define callsym		0x4000000	// 67108864
#define constsym	0x8000000	// 134217728
#define varsym		0x10000000	// 268435456
#define procsym		0x20000000	// 536870912

enum object {
    constant, variable, proc
};

enum fct {
    lit, opr, lod, sto, cal, Int, jmp, jpc	// functions
};

typedef struct{
    enum fct f;		// function code
    long l;		// level
    long a;		// displacement address
} instruction;
/*  lit 0, a : load constant a
    opr 0, a : execute operation a
    lod l, a : load variable l, a
    sto l, a : store variable l, a
    cal l, a : call procedure a at level l
    Int 0, a : increment t-register by a
    jmp 0, a : jump to a
    jpc 0, a : jump conditional to a       */

extern long sym;		// last symbol read
extern char id[al+1];			// last identifier read
extern long num;			// last number read

extern long ll;			// line length
extern long cc;			// character count

extern char line[81];		// 下标为0的单元不使用。
//@ weak global invariant last_char_in_line : line[ll] == ' ';
extern char ch;			// last character read

extern long kk;
//@ strong global invariant ident_length : 0 <= kk <= al;
extern long err;
//@ strong global invariant error_number : 0 <= err <= errmax;


extern char a[al+1];
//@ weak global invariant last_char_in_a : \is_pstring(a, al) && (\forall long m:[kk..al-1]. a[m] == ' ');

extern instruction code[cxmax+1];
extern long cx;			// code allocation index
/*@ strong global invariant cx_inv: 0 <= cx <= cxmax+1; */

extern char word[norw][al+1];
/*@ predicate reservedWord(char word[norw][al+1]) =
	(\forall long i:[0..norw-1].\is_pstring(word[i], al)) &&
	\string(word[0], al) == "begin     " && \string(word[1], al) == "call      " &&
	\string(word[2], al) == "const     " && \string(word[3], al) == "do        " &&
	\string(word[4], al) == "end       " && \string(word[5], al) == "if        " &&
	\string(word[6], al) == "odd       " && \string(word[7], al) == "procedure " &&
	\string(word[8], al) == "then      " && \string(word[9], al) == "var       " &&
	\string(word[10], al) == "while     " &&
	(\forall long i:[0..9]. \string(word[i], al) < \string(word[i+1], al));
    strong global invariant reserved_word_table:
	 reservedWord(word);
*/
extern long wsym[norw];
/*@ predicate reservedWordToken(long wsym[norw]) =
	wsym[0] == beginsym &&
	wsym[1] == callsym && wsym[2] == constsym && wsym[3] == dosym &&
	wsym[4] == endsym && wsym[5] == ifsym && wsym[6] == oddsym &&
	wsym[7] == procsym && wsym[8] == thensym && wsym[9] == varsym &&
	wsym[10] == whilesym;
    strong global invariant reserved_word_token:
	reservedWordToken(wsym);
*/

extern long ssym[256]; 
/*@ predicate operatorToken(long ssym[256]) =
	(\forall long i:[0..39]. ssym[i] == nul )&&
	ssym[40] == lparen && ssym[41] == rparen && ssym[42] == times && ssym[43] == plus &&
	ssym[44] == comma && ssym[45] == minus && ssym[46] == period && ssym[47] == slash &&
	(\forall long i:[48..58]. ssym[i] == nul) && ssym[59] == slash && ssym[60] == nul &&
	ssym[61] == eql && (\forall long i:[62..255]. ssym[i] == nul);
    strong global invariant reserved_word_token:
	operatorToken(ssym);
*/

extern char mnemonic[8][3+1];
/*@ strong global invariant mnemonic_table:
    (\forall long i:[0..8-1].\is_pstring(mnemonic[i], 3)) &&
    \string(mnemonic[0], 3) == "lit" && \string(mnemonic[1], 3) == "opr" &&
    \string(mnemonic[2], 3) == "lod" && \string(mnemonic[3], 3) == "sto" && \string(mnemonic[4], 3) == "cal" &&
    \string(mnemonic[5], 3) == "int" && \string(mnemonic[6], 3) == "jmp" && \string(mnemonic[7], 3) == "jpc";
*/

extern long declbegsys;
//@ strong global invariant declbegsys_inv: declbegsys == (constsym | varsym | procsym);

extern long statbegsys;
//@ strong global invariant statbegsys_inv: statbegsys == (beginsym | callsym | ifsym | whilesym);

extern long facbegsys;
//@ strong global invariant facbegsys_inv: facbegsys == (ident | number | lparen);

extern long tx;		// current table index
//@ strong global invariant tx_inv: 0 <= tx <= txmax;

struct name_tab {
    char name[al+1];
    enum object kind;
    long val;
    long level;
    long addr;
};
extern struct name_tab table[txmax+1];	//对于原来的代码来说，下面的强变量不变式并不妥当。为了维持这个不变式，在miscellany.c文件的enter函数中增加了一个对level域的赋值语句。
//@ strong global invariant table_inv: \forall long i:[1..tx]. (\is_pstring(table[i].name, al) && 0 <= table[i].level <= levmax);

extern char infilename[80];
extern FILE* infile;

// the following variables for block
extern long dx;			// data allocation index
//@ strong global invariant dx_inv: 0 <= dx <= dxmax+1;
extern long lev;		// current depth of block nesting
//@ strong global invariant level_inv:  0 <= lev <= levmax;

// the following array space for interpreter
#define stacksize 50000
extern long s[stacksize];	// datastore



////////////////////////////////////////////////与文法对应的谓词定义//////////////////////////////////////////////////
/*  PL/0语言的语法
Program		::= Block .                                   //pL0程序由程序体和点（.）构成
Block		::= ConstDecl VarDecl ProcDecSeq Stmt         //程序体由常量声明、变量声明、过程声明序列和语句构成
ConstDecl	::= const ConstDefSeq ; | epsilon             //常量声明或者由const、常量定义序列和分号构成，或者为空
ConstDefSeq	::= ConstDef | ConstDefSeq , ConstDef         //常量定义序列由一个或多个常量定义构成
ConstDef	::= ident = number                            //常量定义由常量名（标识符）、赋值号和常量值构成
VarDecl		::= var VarDecSeq ; | epsilon                 //变量声明或者由var和变量声明序列构成，或者为空
VarDecSeq 	::= ident | VarDecSeq , VarDecSeq		      //变量声明序列由一个或多个变量名（标识符）（中间用逗号分隔）构成	
ProcDecSeq	::= ProcDecSeq ProcDecl | epsilon             //过程声明序列或者由一个或多个过程声明构成，或者为空
ProcDecl 	::= procedure ident ; Block ;                 //过程声明由procedure、过程名（标识符）、分号、程序体和分号构成。
Stmt		::= ident := Exp | call ident | begin StmtSeq end |   //语句由赋值语句、过程调用语句、复合语句、条件语句、循环语句和空语句构成
		    if Cond then Stmt | while Cond do Stmt | epsilon
StmtSeq		::= Stmt | StmtSeq ; Stmt                     //语句序列由一个或多个语句（中间用分号分隔）构成
Cond		::= odd Exp | Exp RelOp Exp                   //条件表达式或者是一元奇数判断表达式，或者是二元关系算符表达式
RelOp		::= = | <> | < | > | <= | >=                  //二元关系算符包括等于、不等于、小于、大于、小于等于、大于等于。
Exp		::= + Term TermSeq | - Term TermSeq | Term TermSeq   //表达式或者由正号、项和项序列构成，或者由负号、项和项序列构成，或者由项和项序列构成
TermSeq		::= TermSeq + Term | TermSeq - Term | epsilon	 //项序列由零个或多个+ Term(- Term)构成。
Term		::= Factor FactorSeq                          //项由因子和因子序列构成              
FactorSeq	::= FactorSeq * Factor | FactorSeq / Factor | epsilon	//因子序列由零个或多个* Factor(/ Factor)构成。
Factor		::= ident | number | ( Exp )              //因子或者是标识符，或者是常量值，或者是带括号的表达式
*/

//@ ghost typedef long symArray[];    // 没有指定大小的幽灵数组类型symArray
//@ ghost extern symArray symStream;  // 没有指定大小的幽灵数组symStream记录词法分析产生的单词
//@ ghost extern integer symStreamIndex;

//last == first-1对应文法中的空选择epsilon。对于文法中的epsilon，没有在symStream数组中增加任何符号。因为first和last分别是针对一个非终结符的第1个和最后1个符号的下标，
//那么last == first-1就在情理之中了。而且这个last就是和它相邻的左边非终结符最后一个符号在symStream。
/*@   
inductive Exp(symArray symStream, integer first, integer last);
inductive Factor(symArray symStream, integer first, integer last) =
    last == first && symStream[first] == ident                           //因子是标识符
	||                        
    last == first && symStream[first] == number                          //因子是常量值
	||                       
    last >= first+2 && symStream[first] == lparen && Exp(symStream, first+1, last-1) && symStream[last] == rparen;      //因子是带括号的表达式
inductive FactorSeq(symArray symStream, integer first, integer last) =
    last == first-1                                     //因子序列为空
	||                                                  
    last >= first+1 && (\exists integer middle:[first..last-1]. FactorSeq(symStream, first, middle-1)                   //因子序列由一个或多个* Factor(/ Factor)构成
	&& (symStream[middle] == times || symStream[middle] == slash) && Factor(symStream, middle+1, last));
inductive Term(symArray symStream, integer first, integer last) =
    last >= first && (\exists integer middle:[first..last]. Factor(symStream, first, middle) && FactorSeq(symStream, middle+1, last));   //项由因子和因子序列构成
inductive TermSeq(symArray symStream, integer first, integer last) =
    last == first-1                                    //项序列为空 
	||         
    last >= first+1 && (\exists integer middle:[first..last-1].TermSeq(symStream, first, middle-1)                      //项序列一个或多个+ Term(- Term)构成
	&& (symStream[middle] == plus || symStream[middle] == minus) && Term(symStream, middle+1, last));
inductive Exp(symArray symStream, integer first, integer last) =
    last >= first+1 && (symStream[first] == plus || symStream[first] == minus) &&     //表达式或者由正号、项和项序列构成，或者由负号、项和项序列构成
	(\exists integer middle:[first+1..last]. Term(symStream, first+1, middle) && TermSeq(symStream, middle+1, last)) 
	||
    last >= first && symStream[first] != plus && symStream[first] != minus &&     //表达式由项和项序列构成
	(\exists integer middle:[first..last]. Term(symStream, first, middle) && TermSeq(symStream, middle+1, last));
predicate RelOp(long sym) =
    sym == eql      //等于
	||
	sym == neq      //不等于
	||
	sym == lss      //小于
	||
	sym == leq      //小于等于
	||
	sym == gtr      //大于
	||
	sym == geq;     //大于等于
predicate Cond(symArray symStream, integer first, integer last) =
    last >= first+1 && symStream[first] == oddsym && Exp(symStream, first+1, last)    //一元奇数判断表达式
	||
    last >= first+2 && symStream[first] != oddsym && (\exists integer middle:[first+1..last-1].    //二元关系算符表达式
	Exp(symStream, first, middle-1) && RelOp(symStream[middle]) && Exp(symStream, middle+1, last));
inductive StmtSeq(symArray symStream, integer first, integer last);
inductive Stmt(symArray symStream, integer first, integer last) =
    last == first-1                                     // 空语句
	||													
    last >= first+2 &&symStream[first] == ident && symStream[first+1] == becomes && Exp(symStream, first+2, last)    // 赋值语句
	||	
    last == first+1 && symStream[first] == callsym && symStream[last] == ident                              // 调用语句
	||					
    last >= first+2 && symStream[first] == ifsym && (\exists integer middle:[first+2..last].				// 条件语句
	Cond(symStream, first+1, middle-1) && symStream[middle] == thensym && Stmt(symStream, middle+1, last)) 
	||
    last >= first+2 && symStream[first] == whilesym && (\exists integer middle:[first+2..last].				// 循环语句
	Cond(symStream, first+1, middle-1) && symStream[middle] == dosym && Stmt(symStream, middle+1, last))
	||
    symStream[first] == beginsym && StmtSeq(symStream, first+1, last-1) && symStream[last] == endsym && last >= first+1;// 复合语句
inductive StmtSeq(symArray symStream, integer first, integer last) =
    last == first-1                                      //语句序列为空                           
	||
    last >= first+1 && Stmt(symStream, first, last)      //语句序列为一条非空语句
	||
    last >= first && (\exists integer middle:[first..last].
	StmtSeq(symStream, first, middle-1) && symStream[middle] == semicolon && Stmt(symStream, middle+1, last));    //语句序列为多条语句？？？？？
predicate ConstDef(symArray symStream, integer first, integer last) =
    last == first+2 && symStream[first] == ident && symStream[first+1] == eql && symStream[first+2] == number;   //常量定义由常量名（标识符）、赋值号和常量值构成
inductive ConstDefSeq(symArray symStream, integer first, integer last) =
    last == first+2 && ConstDef(symStream, first, last)                                                          //常量定义序列由一个常量定义构成
	||
    last >= first+6 && ConstDefSeq(symStream, first, last-4) && symStream[last-3] == comma && ConstDef(symStream, last-2, last);  //常量定义序列由多个常量定义构成
predicate ConstDecl(symArray symStream, integer first, integer last) =
    last == first-1                                           //常量声明为空
	||
    last >= first+4 && symStream[first] == constsym && ConstDefSeq(symStream, first+1, last-1) && symStream[last] == semicolon;   //常量声明由const、常量定义序列和分号构成
inductive VarDecSeq(symArray symStream, integer first, integer last) =
    last == first && symStream[first] == ident   //变量声明序列或者由一个变量名（标识符）构成
	||
    last >= first+2 && VarDecSeq(symStream, first, last-2) && symStream[last-1] == comma && VarDecSeq(symStream, last, last); //变量声明序列或者由两个以上变量名（标识符）（中间用逗号分隔）构成
predicate VarDecl(symArray symStream, integer first, integer last) =
    last == first-1  //变量声明或者为空  ？？？？？？                                                   
	||
    last >= first+2 && symStream[first] == varsym && VarDecSeq(symStream, first+1, last-1) && symStream[last] == semicolon; //变量声明或者由var和变量声明序列构成
inductive Block(symArray symStream, integer first, integer last);
inductive ProcDecl(symArray symStream, integer first, integer last) =        
    last >= first+3 && symStream[first] == procsym && symStream[first+1] == ident  //过程声明由procedure、过程名（标识符）、分号、程序体和分号构成。
	&& symStream[first+2] == semicolon && Block(symStream, first+3, last-1) && symStream[last] == semicolon;
inductive ProcDecSeq(symArray symStream, integer first, integer last) =
    last == first-1                                                    //过程声明序列为空
	||
    last >= first+3 && (\exists integer middle:[first..last]. ProcDecSeq(symStream, first, middle-1) && ProcDecl(symStream, middle, last)); //过程声明序列或者由一个或多个过程声明构成
inductive Block(symArray symStream, integer first, integer last) =
    last >= first-1 && (\exists integer middle1:[first-1..last]. \exists integer middle2:[first-1..last].  //程序体由常量声明、变量声明、过程声明序列和语句构成
	\exists integer middle3:[first-1..last]. ConstDecl(symStream, first, middle1) && VarDecl(symStream, middle1+1, middle2) 
	&& ProcDecSeq(symStream, middle2+1, middle3) && Stmt(symStream, middle3+1, last));
predicate Program(symArray symStream, integer first, integer last) =
    last >= first && Block(symStream, first, last-1) && symStream[last] == period;     //pL0程序由程序体和点（.）构成
*/

//////////////////////////////////////////////////与中间代码对应的谓词定义//////////////////////////////////////////////////////////////////////
//中间代码由三个部分构成：f，l，a。f表示功能码，l表示层次差，a用于区别不同的指令（也可表示常量或相对地址）。
//定义各种语法构造的中间代码结构的谓词，忽略了其中变量相对地址和常量值。
//last == first-1也是对应文法中的空选择。和语法分析不同的是，对于代码生成而言，若是空选择，它根本不调用gen函数，不会改变Code数组中的内容，因而可以忽略这个选项。
//为避免忽略空语句选择引起多个函数的标注需要修改，我们暂不做修改。

/*@ 
inductive ExpCode(instruction code[cxmax+1], long first, long last);
inductive FactorCode(instruction code[cxmax+1], long first, long last) =
    last == first && code[first].f == lit && code[first].l == 0      //因子是常量，对应的中间代码是LIT 0 a
	||
    last == first && code[first].f == lod                            //因子是变量标识符，对应的中间代码是LOD l a
	||
    last >= first && ExpCode(code, first, last);                     //因子是带括号的表达式
inductive FactorSeqCode(instruction code[cxmax+1], long first, long last) =
    last == first-1                                                  //因子序列为空。
	||                                                               
    last >= first+1 && (\exists long middle:[first..last-1].	     //因子序列由多个* Factor(/ Factor)构成，对应的最后一条中间代码是OPR 0 4或OPR 0 5
	FactorSeqCode(code, first, middle-1) && FactorCode(code, middle, last-1) && code[last].f == opr &&
	    code[last].l == 0 && (code[last].a == 4 || code[last].a == 5));
inductive TermCode(instruction code[cxmax+1], long first, long last) =  //项由因子和因子序列构成，对应的中间代码由因子和因子序列翻译得到
    last >= first && (\exists long middle:[first..last].FactorCode(code, first, middle) && FactorSeqCode(code, middle+1, last));
inductive TermSeqCode(instruction code[cxmax+1], long first, long last) =
    last == first-1                                                         //项序列为空。
	||                                                 
    last >= first+1 && (\exists long middle:[first..last-1].              //项序列由多个+ Term(- Term)构成，对应的最后一条中间代码是OPR 0 2或OPR 0 3     
	TermSeqCode(code, first, middle-1) && TermCode(code, middle, last-1) && code[last].f == opr &&
	    code[last].l == 0 && (code[last].a == 2 || code[last].a == 3));
inductive ExpCode(instruction code[cxmax+1], long first, long last) = 
    last >= first+1 && (\exists long middle:[first+1..last].(TermCode(code, first, middle-1)
	&& code[middle].f == opr && code[middle].l == 0 && code[middle].a == 1 && TermSeqCode(code, middle+1, last)))   //表达式或者由负号、项和项序列构成，对应的中间代码中包含OPR 0 1
	||
    last >= first && (\exists long middle:[first..last].TermCode(code, first, middle) && TermSeqCode(code, middle+1, last)); //表达式或者由正号、项和项序列构成，或者由项和项序列构成
predicate CondCode(instruction code[cxmax+1], long first, long last) = 
    last >= first+2     //条件表达式是二元关系算符表达式，对应的最后一条中间代码是OPR 0 8(或9、10、11、12、13)  
	&& (\exists long middle:[first..last-1].ExpCode(code, first, middle) && ExpCode(code, middle+1, last-1))	
	&& code[last].f == opr && code[last].l == 0 
	&& (code[last].a == 8 || code[last].a == 9 || code[last].a == 10 || code[last].a == 11 || code[last].a == 12 || code[last].a == 13) 
	||
    last >= first+1 && ExpCode(code, first, last-1) && code[last].f == opr && code[last].l == 0 && code[last].a == 6;//条件表达式是一元奇数判断表达式，对应的最后一条中间代码是OPR 0 6   
inductive StmtSeqCode(instruction code[cxmax+1], long first, long last);
inductive StmtCode(instruction code[cxmax+1], long first, long last) =
    last == first-1                                                             // 语句为空
	||														
    last >= first+1 && ExpCode(code, first, last-1) && code[last].f == sto      //赋值语句对应的最后一条中间代码的功能码是STO 
	||							
    last == first && code[first].f == cal                                       // 过程调用语句对应的最后一条中间代码的功能码是CAL
	||											
    last >= first+2 && (\exists long middle:[first..last-1].(CondCode(code, first, middle)    //条件语句对应的中间代码中包含 jpc 0  a(if语句下一条语句的起始地址)
	&& code[middle+1].f == jpc && code[middle+1].l == 0 && code[middle+1].a == last+1 && StmtCode(code, middle+2, last))) 
	||
    last >= first+3 && (\exists long middle:[first..last-2].(CondCode(code, first, middle)   //循环语句对应的中间代码中包含 jpc 0  a(循环语句下一条语句的起始地址)
	&& code[middle+1].f == jpc && code[middle+1].l == 0 && code[middle+1].a == last+1 && StmtCode(code, middle+2, last-1))
	&& code[last].f == jmp && code[last].l == 0 && code[last].a == first) 
	||
    last >= first && StmtSeqCode(code, first, last);										// 语句序列
inductive StmtSeqCode(instruction code[cxmax+1], long first, long last) =
    last >= first-1 && StmtCode(code, first, last)      //语句序列由一个语句构成
	||
    last >= first && (\exists long middle:[first..last].StmtSeqCode(code, first, middle) && StmtCode(code, middle+1, last)); //语句序列由两个或多个语句构成
inductive BlockCode(instruction code[cxmax+1], long first, long last);
inductive ProcCode(instruction code[cxmax+1], long first, long last) = last >= first+1 && BlockCode(code, first, last);	//过程声明对应的中间代码由该声明自己的程序体翻译得到 
inductive ProcSeqCode(instruction code[cxmax+1], long first, long last) =
    last == first-1                        //过程声明序列为空
	||
    last>=first+2 && (\exists integer middle:[first-1..last-2].ProcSeqCode(code, first, middle) && ProcCode(code, middle+1, last));//过程声明序列由一个或多个过程声明构成
inductive BlockCode(instruction code[cxmax+1], long first, long last) =
    last >= first+2 && code[first].f == jmp && code[first].l == 0  //程序体对应的中间代码包含jmp指令、INT指令和OPR 0 0调用返回指令
	&&
    (\exists long middle:[first..last-1]. code[first].a == middle+1 && ProcSeqCode(code, first+1, middle)
	&& code[middle+1].f == Int && code[middle+1].l == 0 && StmtCode(code, middle+2, last-1))
	&& code[last].f == opr && code[last].l == 0 && code[last].a == 0;
predicate ProgramCode(instruction code[cxmax+1], long first, long last) = last >= first+1 && BlockCode(code, first, last); //PL0程序对应的中间代码由程序体翻译得到
*/

///////////////////////////////////////全局逻辑定义和声明/////////////////////////////////////////////
/*@ 
	//逻辑函数，计算m的n次幂
	logic long power(long m,long n) = 
		n == 0 ? 1 : m * power(m, n-1);
*/

/*@
	//谓词，描述形参sym所有可能取值
	predicate currentSymbolFunc(long sym, char id[al+1], long num)=
		sym == beginsym || sym == endsym || sym == ifsym || sym == thensym 
		|| sym == whilesym || sym == dosym || sym == callsym || sym == constsym
		|| sym == varsym || sym == procsym || sym == oddsym || sym == ident && \is_pstring(id, al)
		|| sym == number && 0 <= num < power(10, nmax) || sym == neq || sym == lss || sym == leq 
		|| sym == gtr || sym == geq || sym==becomes || sym == plus || sym == minus || sym == times
		|| sym == slash || sym == lparen || sym == rparen || sym == comma || sym == semicolon 
		|| sym == period || sym == eql || sym == nul
		;
*/

/*@ 
	//宏，描述词法分析所用缓冲区相关全局变量的取值
	define currentCodeLine: 0 < cc <= 80 && 0 < ll <= 80 && cc <= ll &&
		(\forall long i:[1..ll-1].0 <= line[i] <= 255) && ch == line[cc] && line[ll] == ' ' 
		&& \length(infile) == 1 && \offset(infile) == 0 && (infile->_flags & _IO_NO_READS) == 0
		;
	define currentCodeLineAndExtra: currentCodeLine && 0 <= cx <= cxmax+1 && \is_pstring(a, al) && (\forall long m:[kk..al-1]. a[m] == ' ');
	//宏，描述当前记号sym所有可能取值
	define currentSymbol:currentSymbolFunc(sym, id, num);
*/

////////////////////////////////////////////////辅助谓词定义//////////////////////////////////////////////////
/*@
	//谓词，描述形参ch是字符
	predicate is_alpha(char ch) = 
		65 <= ch <= 90 || 97 <= ch <= 122
		;
	//谓词，描述形参ch是数字
	predicate is_digit(char ch) =
		48 <= ch <= 57
		;   
*/

/*@
	//逻辑函数，计算数字数组段对应的的整数值
	logic long a_to_i(char t[81], long first, long k) =
		(k == 0) ? 0 : a_to_i(t, first, k-1)*10 + t[first+k-1] - '0';
*/

/*@ 
ghost extern _Bool too_many_error, stop_here, nested_level_too_shallow;
*/

/////////////////////////////////////error函数的验证///////////////////////////////////////
//error函数主要功能是输出错误信息，验证的性质都是平凡的；
/*@
	requires
		0 < cc <= ll && cc <= 80 && ll <= 80
		;
	assigns err, too_many_error;
	ensures
		0 < cc <= ll
		;
	exits
		\exit_status == 1 && too_many_error == true
		;
*/
void error(const long n);


//@ ghost extern _Bool program_incomplete, program_line_overflow, too_many_variables;

/////////////////////////////////////getch函数的验证///////////////////////////////////////
//getch函数验证的性质：
//getch执行结束时，全局变量ch的值是缓冲区line中下标为cc的元素，
//当前行结束时，cc是新的一行的起始字符的下标，即1
//当前行还有剩余字符，cc用于获取下一个字符，即cc == \old(cc)+1     
/*@												// 暂时用这个多命名行为的函数协议。
	requires
		0 <= cc <= 80 && 0 <= ll <=80 && cc <= ll && \old(cc) == cc && \old(ll) == ll 
		&& \old(line[0..80]) == line[0..80] && \old(program_incomplete) == program_incomplete
		&& \old(program_line_overflow) == program_line_overflow 
		&& \length(infile) == 1 && \offset(infile) == 0 && (infile->_flags & _IO_NO_READS) == 0 && (cc == ll || cc > 0)
		;
	assigns cc, ch;	    
	ensures 
		currentCodeLine && ch == line[cc]
		;
	behavior new_line:							// 当前行已结束，读入新的一行。
    assumes 
		\old(cc) == \old(ll)
		;
    requires
		\true
		;
    assigns ll, line[0..80], infile->_flags, program_incomplete, program_line_overflow;
    ensures
	cc == 1
	;
    exits
		(cc == ll && program_incomplete == true || ll == 80 && program_line_overflow == true || ch == -1)
		&& \exit_status == 1
		;
	behavior current_line:						// 当前行还有剩余字符。
    assumes
		0 < \old(cc) < \old(ll)
		;
    requires
		currentCodeLine
		;
    ensures
		1 < cc <= ll && cc == \old(cc)+1 && \old(line[0..80]) == line[0..80] && \old(ll) == ll
		&& \old(program_incomplete) == program_incomplete && \old(program_line_overflow) == program_line_overflow
		;
	disjoint behaviors;
	complete behaviors;
*/
// 增加了检查输入数据每行不超过80个字符的代码和代码和断言，以保证line数组不会溢出。
void getch();

//@ ghost extern _Bool program_too_long;

/////////////////////////////////////getsym函数的验证///////////////////////////////////////
//getsym函数验证的性质：
//getsym执行结束时，全局变量sym的值是下一个单词（幽灵数组symStream中下标为\old(symStreamIndex) + 1的元素）。
/*@
	requires
		\old(symStreamIndex) == symStreamIndex && symStreamIndex >= -1 && currentCodeLine && ch == line[cc]
		;
    assigns sym, a, id, kk, num, program_too_long, symStreamIndex, symStream[\old(symStreamIndex)+1], cc, ll, line, ch, err,infile->_flags, 
	    program_incomplete, program_line_overflow, too_many_error;	//从cc开始，是间接被赋值的外部变量。
    ensures
		symStreamIndex >= 0 && symStreamIndex == \old(symStreamIndex) + 1 && symStream[symStreamIndex] == sym && currentCodeLine && currentSymbol
		;
    exits
		\exit_status == 1 && program_too_long == true
		;
*/
void getsym();

/////////////////////////////////////gen函数的验证///////////////////////////////////////
//gen函数验证的性质：
//gen执行结束时，code数组中增加一条中间代码，该中间代码的f，l，a分别等于三个实参。
/*@
	requires
		currentCodeLine && \old(cx) == cx && (\forall long i:[0..\old(cx)-1].\old(code[i]) == code[i])
		;
	assigns cx, code[\old(cx)], program_too_long;		// cx有强类型不变式。
	ensures
		currentCodeLine && cx == \old(cx)+1 && (\forall long i:[0..\old(cx)-1].\old(code[i]) == code[i])
		&& code[\old(cx)].f == x && code[\old(cx)].l == y && code[\old(cx)].a == z
		;
	exits
		\exit_status == 1 && program_too_long == true
		;
*/
void gen(const enum fct x, const long y, const long z);

/////////////////////////////////////test函数的验证///////////////////////////////////////
//test函数验证的性质：
//sym在集合s1中，则函数执行结束时，sym和symStreamIndex保持不变；
//sym不在集合s1中，则函数执行结束时，sym在集合s1或s2中，sym在幽灵数组symStream中的下标保持不变
//
/*@										
	requires
		currentCodeLineAndExtra && \old(s1) == s1 && \old(err) == err && \old(too_many_error) == too_many_error 
		&& \old(stop_here) == stop_here && \old(sym)== sym
		;
	behavior success:
    assumes
		(\old(sym) & \old(s1)) != 0
		;
    ensures
		currentCodeLineAndExtra && \old(err) == err && \old(too_many_error) == too_many_error 
		&& \old(stop_here) == stop_here && \old(sym)== sym
		;
	behavior failure:
    assumes
		(\old(sym) & \old(s1)) == 0
		;
    assigns err, too_many_error, stop_here;	  
    ensures
		\true
		; 					//currentCodeLineAndExtra;
    exits
		\exit_status == 2 && stop_here == true
		; 	
  complete behaviors;
  disjoint behaviors;
*/
void test(long s1, long const s2, long n); 


//@ ghost extern _Bool symbol_table_too_short;

////////////////////////////////////enter函数的验证///////////////////////////////////////
//enter函数验证的性质：
//根据形参的取值，符号表收集不同的信息
//当形参取值constant，符号表收集常量名对应的常量值；
//当形参取值variable，符号表收集变量名对应的层次和偏移地址；
//当形参取值proc，符号表收集过程名对应的层次；
/*@											// 用多命名行为的函数协议。
requires currentCodeLine && \is_pstring(id, al) && \old(tx) == tx &&			// table有强类型不变式，tx有强数据不变式。
	\old(lev) == lev && \old(dx) == dx && \old(too_many_variables) == too_many_variables &&
	\forall long i:[1..tx].\string(table[i].name, al) == \old(\string(table[i].name, al));
assigns tx, err, table[\old(tx)+1], symbol_table_too_short, too_many_error;
ensures currentCodeLine && table[tx].kind == k &&  
    (\forall long i:[1..tx-1].\string(table[i].name, al) == \old(\string(table[i].name, al))) && \string(table[tx].name, al) == \string(id, al);
exits \exit_status == 1 && symbol_table_too_short == true;
  behavior constant_id:
    assumes k == constant;
    ensures table[tx].val == num && \old(lev) == lev && \old(dx) == dx && tx == \old(tx)+1 && \old(too_many_variables) == too_many_variables;
  behavior variable_id:
    assumes k == variable;
    assigns dx, too_many_variables;
    ensures table[tx].level == lev && table[tx].addr == dx-1 && tx == \old(tx)+1;
    exits \exit_status == 1 && too_many_variables == true;
  behavior proc_id:
    assumes k == proc;
    requires \old(dx) == dx;
    ensures table[tx].level == lev && \old(dx) == dx && tx == \old(tx)+1 && \old(too_many_variables) == too_many_variables;
  complete behaviors;									// 正交性的证明依据k的类型。
  disjoint behaviors;

*/
void enter(const enum object k);

////////////////////////////////////position函数的验证///////////////////////////////////////
//position函数验证的性质：
//当函数返回值大于0时，该返回值是形参id1在符号表中条目的索引
//当函数返回值等于0时，形参id1在符号表中无相应条目
/*@
requires \is_pstring(id1, al) && \forall long i:[1..tx].\is_pstring(table[i].name, al) && \string(table[i].name, al) == \old(\string(table[i].name, al));
assigns table[0].name;
ensures 1 <= \result <= tx && \is_pstring(id1, al) && (\forall long i:[1..tx].\is_pstring(table[i].name, al) && \string(table[i].name, al) == \old(\string(table[i].name, al))) &&
	    \string(table[\result].name, al) == \string(id1, al) ||
	\result == 0 && \is_pstring(id1, al) && (\forall long i:[1..tx].\is_pstring(table[i].name, al) && \string(table[i].name, al) == \old(\string(table[i].name, al))) &&
	    (\forall long n:[1.. tx]. \string(table[n].name, al) != \string(id1, al));
*/
long position(const char* const id1);

////////////////////////////////////constdeclaration函数的验证///////////////////////////////////////
//constdeclaration函数验证的性质：
//单词流符合常量声明的语法规则
//符号表收集了常量名的各种属性
/*@
	requires
		currentCodeLineAndExtra && currentSymbol && \old(symStreamIndex) == symStreamIndex && symStreamIndex >= 0
		&& symStream[symStreamIndex] == sym && \old(tx) == tx
		;
	assigns	num, stop_here, a, id, kk, tx, table[\old(tx)+1], cc, ll, err, sym, symStreamIndex, symStream[\old(symStreamIndex)+1..\old(symStreamIndex)+3],infile->_flags,
	line, ch, program_incomplete, program_line_overflow, program_too_long, symbol_table_too_short, lev, too_many_error;
	ensures
		currentCodeLineAndExtra && currentSymbol && ConstDef(symStream, \old(symStreamIndex), symStreamIndex-1)
		&& symStream[symStreamIndex] == sym && symStreamIndex >= 0 && symStreamIndex == \old(symStreamIndex)+3
		&& \string(table[tx].name, al) == \string(id, al) && table[tx].kind == constant && table[tx].val == num
		&& 0 <= num < power(10, nmax) && tx == \old(tx)+1
		;
	exits
		\exit_status == 2 && stop_here == true
		;
*/
void constdeclaration();

////////////////////////////////////vardeclaration函数的验证///////////////////////////////////////
//vardeclaration函数验证的性质：
//单词流符合变量声明的语法规则
//符号表中记录了变量名的各种属性
/*@
	requires
		currentCodeLineAndExtra && currentSymbol && \old(symStreamIndex) == symStreamIndex && symStreamIndex >= 0
		&& symStream[symStreamIndex] == sym && \old(tx) == tx
		;
	assigns	stop_here, too_many_variables, a, id, kk, tx, table[\old(tx)+1], cc, ll, err, sym, symStreamIndex, symStream[\old(symStreamIndex)+1],infile->_flags,
	num, line, ch, program_incomplete, program_line_overflow, too_many_variables, program_too_long, symbol_table_too_short, dx, lev, too_many_error;
	ensures
		currentCodeLineAndExtra && currentSymbol && VarDecSeq(symStream, \old(symStreamIndex), symStreamIndex - 1)
		&& symStreamIndex >= 0 && symStreamIndex == \old(symStreamIndex) + 1 && symStream[symStreamIndex] == sym
		&& \string(table[tx].name, al) == \string(id, al) && table[tx].kind == variable
		&& table[tx].level == lev && table[tx].addr == dx-1 && tx > \old(tx)
		;
	exits
		\exit_status == 2 && stop_here == true
		;
*/
void vardeclaration();

/*@
	requires
		0 <= cx0 < cx
		;
	ensures
		0 <= cx0 < cx
		;
*/
void listcode(const long cx0);

////////////////////////////////////expression函数的验证///////////////////////////////////////
//expression函数验证的性质：
//单词流符合表达式的语法规则
//该函数执行时生成的中间代码结构正确
/*@
	requires
		currentCodeLineAndExtra && currentSymbol && \old(symStreamIndex) == symStreamIndex && symStreamIndex >= 0 
		&& symStream[symStreamIndex] == sym && \old(cx) == cx
		;
	assigns	a, id, kk, cx, cc, ll, err, sym, symStreamIndex, symStream[\old(symStreamIndex)+1..\infinity], code[\old(cx)..cxmax], table[0].name,infile->_flags,
		num, line, ch, program_incomplete, program_line_overflow, program_too_long, stop_here, too_many_error;
	ensures
		currentCodeLineAndExtra && currentSymbol && Exp(symStream, \old(symStreamIndex), symStreamIndex-1) && symStreamIndex > \old(symStreamIndex) >= 0
		&& symStream[symStreamIndex] == sym && ExpCode(code, \old(cx), cx-1) && cx > \old(cx)
		;
*/
void expression(const long fsys);

////////////////////////////////////factor函数的验证///////////////////////////////////////
//factor函数验证的性质：
//单词流符合因子的语法规则
//该函数执行时生成的中间代码结构正确
/*@
	requires
		currentCodeLineAndExtra && currentSymbol && \old(symStreamIndex) == symStreamIndex && symStreamIndex >= 0 
		&& symStream[symStreamIndex] == sym && \old(cx) == cx
		;
	assigns	num, stop_here, a, id, kk, cx, cc, ll, err, sym, symStreamIndex, symStream[\old(symStreamIndex)+1..\infinity], code[\old(cx)..cxmax],infile->_flags,
		line,ch,program_incomplete, program_line_overflow, program_too_long,table[0].name, too_many_error;
	ensures
		currentCodeLineAndExtra && currentSymbol && symStream[symStreamIndex] == sym 
		&& symStreamIndex > \old(symStreamIndex) >= 0 && Factor(symStream, \old(symStreamIndex), symStreamIndex-1)
		&& FactorCode(code, \old(cx), cx-1) && cx > \old(cx)
		;
	exits
		\exit_status == 2 && stop_here == true || \exit_status == 1 && program_too_long == true
		;
*/
void factor(const long fsys);

////////////////////////////////////term函数的验证///////////////////////////////////////
//term函数验证的性质：
//单词流是否符合项的语法规则
//该函数执行时生成的中间代码结构是否正确
/*@
	requires
		currentCodeLineAndExtra && currentSymbol && \old(symStreamIndex) == symStreamIndex && symStreamIndex >= 0 
		&& symStream[symStreamIndex] == sym && \old(cx) == cx
		;
	assigns	a, id, kk, cx, cc, ll, err, sym, symStreamIndex, symStream[\old(symStreamIndex)+1..\infinity], code[\old(cx)..cxmax], num, line, ch, program_incomplete,infile->_flags,
        program_line_overflow, program_too_long, stop_here, table[0].name, too_many_error;
	ensures
		currentCodeLineAndExtra && currentSymbol && symStream[symStreamIndex] == sym 
		&& symStreamIndex > \old(symStreamIndex) >= 0 && Term(symStream, \old(symStreamIndex), symStreamIndex-1)
		&& TermCode(code, \old(cx), cx-1) && cx > \old(cx)
		;
*/
void term(const long fsys);

////////////////////////////////////condition函数的验证///////////////////////////////////////
//condition函数验证的性质：
//单词流符合条件表达式的语法规则
//该函数执行时生成的中间代码结构正确
/*@
	requires
		currentCodeLineAndExtra && currentSymbol && \old(symStreamIndex) == symStreamIndex && symStreamIndex >= 0
		&& symStream[symStreamIndex] == sym && \old(cx) == cx
		;
	assigns	stop_here, a, id, kk, cx, cc, ll, err, sym, symStreamIndex, symStream[\old(symStreamIndex)+1..\infinity], code[\old(cx)..cxmax],infile->_flags,
		num,line,ch,program_incomplete, program_line_overflow, table[0].name, program_too_long,too_many_error;
	ensures
		currentCodeLineAndExtra && currentSymbol && Cond(symStream, \old(symStreamIndex), symStreamIndex-1) && symStream[symStreamIndex] == sym
		&& CondCode(code, \old(cx), cx-1) && symStreamIndex > \old(symStreamIndex) >= 0 && cx > \old(cx)
		;
	exits
		\exit_status == 2 && stop_here == true
		;
*/
void condition(const long fsys);

/////////////////////////////////////statement函数的验证///////////////////////////////////////
//statement函数验证的性质：
//单词流符合语句的语法规则
//该函数执行时生成的中间代码结构正确
/*@
	requires
		currentCodeLineAndExtra && currentSymbol && \old(symStreamIndex) == symStreamIndex && symStreamIndex >= 0
		&& symStream[symStreamIndex] == sym && \old(cx) == cx
		;
	assigns stop_here, program_too_long, a, id, kk, cx, cc, ll, err, sym, symStreamIndex, symStream[\old(symStreamIndex)+1..\infinity], code[\old(cx)..cxmax],infile->_flags,
		num, line, ch, program_incomplete, program_line_overflow, table[0].name, too_many_error;
	ensures
		currentCodeLineAndExtra && currentSymbol && Stmt(symStream, \old(symStreamIndex), symStreamIndex-1) && StmtCode(code, \old(cx), cx-1)
		&& symStream[symStreamIndex] == sym && symStreamIndex >= \old(symStreamIndex) >= 0 && cx >= \old(cx)
		;
	exits
		\exit_status == 2 && stop_here == true || \exit_status == 1 && program_too_long == true
		;
*/
void statement(const long fsys);

/////////////////////////////////////block函数的验证///////////////////////////////////////
//block函数验证的性质：
//单词流符合程序块的语法规则
//该函数执行时生成的中间代码结构正确
/*@
	requires
		currentCodeLineAndExtra && currentSymbol && \old(symStreamIndex) == symStreamIndex && symStreamIndex >= 0 
		&& symStream[symStreamIndex] == sym && \old(cx) == cx && \old(tx) == tx
		;
	assigns stop_here, nested_level_too_shallow, dx, cx, tx, table[\old(tx)].addr, table[\old(tx)+1..txmax], lev,infile->_flags, 
		a, id, kk, cc, ll, err, sym, symStreamIndex, symStream[\old(symStreamIndex)+1..\infinity], code[\old(cx)..cxmax],
		num, line, ch, program_incomplete, program_line_overflow, too_many_variables, program_too_long, symbol_table_too_short, too_many_error, table[0].name;
	ensures
		currentCodeLineAndExtra && currentSymbol && Block(symStream, \old(symStreamIndex), symStreamIndex-1) && BlockCode(code, \old(cx), cx-1)
		&& symStream[symStreamIndex] == sym && symStreamIndex >= \old(symStreamIndex) >= 0 && cx > \old(cx)
		;
	exits
		\exit_status == 1 && nested_level_too_shallow == true || \exit_status == 2 && stop_here == true
		;
*/
void block(long fsys);



//@ ghost _Bool program_too_long;

void gen(const enum fct x, const long y, const long z){
    if(cx > cxmax){
		printf("%s\n", "program too long");
		//@ ghost program_too_long = true;
		exit(1);
    }
    code[cx].f=x; code[cx].l=y; code[cx].a=z;
    cx=cx+1;
}
